@@ -27,7 +27,7 @@
 #include <QtCore/QBitArray>
 #include <QtCore/QStack>
 
-#include <QtGui/private/qrhi_p.h>
+#include <rhi/qrhi.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -115,7 +115,7 @@ public:
         // one. when an item is released, we'll reset m_freePage anyway.
         if (!p) {
             p = new AllocatorPage<Type, PageSize>();
-            m_freePage = pages.count();
+            m_freePage = pages.size();
             pages.push_back(p);
         }
         uint pos = p->blocks[PageSize - p->available];
@@ -259,7 +259,7 @@ inline QDebug operator << (QDebug d, const Rect &r) {
 }
 
 struct Buffer {
-    int size;
+    quint32 size;
     // Data is only valid while preparing the upload. Exception is if we are using the
     // broken IBO workaround or we are using a visualization mode.
     char *data;
@@ -597,6 +597,10 @@ struct GraphicsState
     bool blending = false;
     QRhiGraphicsPipeline::BlendFactor srcColor = QRhiGraphicsPipeline::One;
     QRhiGraphicsPipeline::BlendFactor dstColor = QRhiGraphicsPipeline::OneMinusSrcAlpha;
+    QRhiGraphicsPipeline::BlendFactor srcAlpha = QRhiGraphicsPipeline::One;
+    QRhiGraphicsPipeline::BlendFactor dstAlpha = QRhiGraphicsPipeline::OneMinusSrcAlpha;
+    QRhiGraphicsPipeline::BlendOp opColor = QRhiGraphicsPipeline::Add;
+    QRhiGraphicsPipeline::BlendOp opAlpha = QRhiGraphicsPipeline::Add;
     QRhiGraphicsPipeline::ColorMask colorWrite = QRhiGraphicsPipeline::ColorMask(0xF);
     QRhiGraphicsPipeline::CullMode cullMode = QRhiGraphicsPipeline::None;
     bool usesScissor = false;
@@ -605,6 +609,7 @@ struct GraphicsState
     QSGGeometry::DrawingMode drawMode = QSGGeometry::DrawTriangles;
     float lineWidth = 1.0f;
     QRhiGraphicsPipeline::PolygonMode polygonMode = QRhiGraphicsPipeline::Fill;
+    int multiViewCount = 0;
 };
 
 bool operator==(const GraphicsState &a, const GraphicsState &b) noexcept;
@@ -638,16 +643,25 @@ bool operator==(const GraphicsPipelineStateKey &a, const GraphicsPipelineStateKe
 bool operator!=(const GraphicsPipelineStateKey &a, const GraphicsPipelineStateKey &b) noexcept;
 size_t qHash(const GraphicsPipelineStateKey &k, size_t seed = 0) noexcept;
 
+struct ShaderKey
+{
+    QSGMaterialType *type;
+    QSGRendererInterface::RenderMode renderMode;
+    int multiViewCount;
+};
+
+bool operator==(const ShaderKey &a, const ShaderKey &b) noexcept;
+bool operator!=(const ShaderKey &a, const ShaderKey &b) noexcept;
+size_t qHash(const ShaderKey &k, size_t seed = 0) noexcept;
+
 struct ShaderManagerShader
 {
     ~ShaderManagerShader() {
-        delete programRhi.program;
+        delete materialShader;
     }
-    struct {
-        QSGMaterialShader *program = nullptr;
-        QRhiVertexInputLayout inputLayout;
-        QVarLengthArray<QRhiGraphicsShaderStage, 2> shaderStages;
-    } programRhi;
+    QSGMaterialShader *materialShader = nullptr;
+    QRhiVertexInputLayout inputLayout;
+    QVarLengthArray<QRhiShaderStage, 2> stages;
     float lastOpacity;
 };
 
@@ -674,11 +688,16 @@ public Q_SLOTS:
     void invalidated();
 
 public:
-    Shader *prepareMaterial(QSGMaterial *material, const QSGGeometry *geometry = nullptr, QSGRendererInterface::RenderMode renderMode = QSGRendererInterface::RenderMode2D);
-    Shader *prepareMaterialNoRewrite(QSGMaterial *material, const QSGGeometry *geometry = nullptr, QSGRendererInterface::RenderMode renderMode = QSGRendererInterface::RenderMode2D);
+    Shader *prepareMaterial(QSGMaterial *material,
+                            const QSGGeometry *geometry = nullptr,
+                            QSGRendererInterface::RenderMode renderMode = QSGRendererInterface::RenderMode2D,
+                            int multiViewCount = 0);
+    Shader *prepareMaterialNoRewrite(QSGMaterial *material,
+                                     const QSGGeometry *geometry = nullptr,
+                                     QSGRendererInterface::RenderMode renderMode = QSGRendererInterface::RenderMode2D,
+                                     int multiViewCount = 0);
 
 private:
-    typedef QPair<QSGMaterialType *, QSGRendererInterface::RenderMode> ShaderKey;
     QHash<ShaderKey, Shader *> rewrittenShaders;
     QHash<ShaderKey, Shader *> stockShaders;
 
@@ -723,7 +742,7 @@ protected:
     QHash<Node *, uint> m_visualizeChangeSet;
 };
 
-class Q_QUICK_PRIVATE_EXPORT Renderer : public QSGRenderer
+class Q_QUICK_EXPORT Renderer : public QSGRenderer
 {
 public:
     Renderer(QSGDefaultRenderContext *ctx, QSGRendererInterface::RenderMode renderMode = QSGRendererInterface::RenderMode2D);
@@ -777,7 +796,7 @@ private:
     friend class RhiVisualizer;
 
     void destroyGraphicsResources();
-    void map(Buffer *buffer, int size, bool isIndexBuf = false);
+    void map(Buffer *buffer, quint32 byteSize, bool isIndexBuf = false);
     void unmap(Buffer *buffer, bool isIndexBuf = false);
 
     void buildRenderListsFromScratch();
@@ -798,7 +817,8 @@ private:
     bool ensurePipelineState(Element *e, const ShaderManager::Shader *sms, bool depthPostPass = false);
     QRhiTexture *dummyTexture();
     void updateMaterialDynamicData(ShaderManager::Shader *sms, QSGMaterialShader::RenderState &renderState,
-                                   QSGMaterial *material, const Batch *batch, Element *e, int ubufOffset, int ubufRegionSize);
+                                   QSGMaterial *material, const Batch *batch, Element *e, int ubufOffset, int ubufRegionSize,
+                                   char *directUpdatePtr);
     void updateMaterialStaticData(ShaderManager::Shader *sms, QSGMaterialShader::RenderState &renderState,
                                   QSGMaterial *material, Batch *batch, bool *gstateChanged);
     void checkLineWidth(QSGGeometry *g);
@@ -858,6 +878,9 @@ private:
     QDataBuffer<Element *> m_tmpAlphaElements;
     QDataBuffer<Element *> m_tmpOpaqueElements;
 
+    QDataBuffer<QRhiBuffer *> m_vboPool;
+    QDataBuffer<QRhiBuffer *> m_iboPool;
+
     uint m_rebuild;
     qreal m_zRange;
 #if defined(QSGBATCHRENDERER_INVALIDATE_WEDGED_NODES)
